// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using Geekeey.Extensions.Process.Buffered;

namespace Geekeey.Extensions.Process.Tests;

[TestFixture]
internal sealed class BufferedExecuteTests
{
	private static Command Echo(string target, string value) => new Command(Dummy.Program.FilePath)
		.WithArguments(["echo", "--target", target, value]);

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_get_the_stdout()
	{
		// Arrange
		var cmd = Echo("stdout", "Hello stdout");

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.StandardOutput.Trim(), Is.EqualTo("Hello stdout"));
			Assert.That(result.StandardError.Trim(), Is.Empty);
		});
	}

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_get_the_stderr()
	{
		// Arrange
		var cmd = Echo("stderr", "Hello stderr");

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.StandardOutput.Trim(), Is.Empty);
			Assert.That(result.StandardError.Trim(), Is.EqualTo("Hello stderr"));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_get_the_stdout_and_stderr()
	{
		// Arrange
		var cmd = Echo("all", "Hello stdout and stderr");

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.StandardOutput.Trim(), Is.EqualTo("Hello stdout and stderr"));
			Assert.That(result.StandardError.Trim(), Is.EqualTo("Hello stdout and stderr"));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_not_hang_on_large_stdout_and_stderr()
	{
		// Arrange
		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["generate", "clob", "--target", "all", "--length", "100000"]);

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.StandardOutput.Trim(), Is.Not.Empty.Or.Null);
			Assert.That(result.StandardError.Trim(), Is.Not.Empty.Or.Null);
		});
	}
}