// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.Text;

using Geekeey.Extensions.Process.Buffered;

namespace Geekeey.Extensions.Process.Tests;

[TestFixture]
internal sealed class CancellationTests
{
	private static Action<string> NotifyOnStart(out TaskCompletionSource tcs)
	{
		// run the continuation async on the thread pool, to allow the io reader to complte
		var source = tcs = new TaskCompletionSource(TaskCreationOptions.RunContinuationsAsynchronously);
		return line =>
		{
			if (line.Contains("Sleeping for", StringComparison.OrdinalIgnoreCase))
				source.SetResult();
		};
	}

	[Test]
	public async Task I_can_execute_a_command_and_cancel_it_immediately()
	{
		// Arrange
		using var cts = new CancellationTokenSource();

		var stdout = new StringBuilder();

		var target = PipeTarget.Merge(
			PipeTarget.ToDelegate(NotifyOnStart(out var tcs)),
			PipeTarget.ToStringBuilder(stdout)
		);

		var cmd = new Command(Dummy.Program.FilePath)
			          .WithArguments(["sleep", "00:00:30"])
		          | target;

		// Act
		var task = cmd.ExecuteAsync(cts.Token);
		await tcs.Task;
		await cts.CancelAsync();

		// Assert
		var exception = Assert.ThrowsAsync<OperationCanceledException>(async () => await task);

		Assert.Multiple(() =>
		{
			Assert.That(ProcessTree.HasExited(task.ProcessId));
			Assert.That(stdout.ToString(), Does.Contain("Sleeping for"));
			Assert.That(stdout.ToString(), Does.Not.Contain("Done."));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_and_kill_it_immediately()
	{
		// Arrange
		var stdout = new StringBuilder();

		var target = PipeTarget.Merge(
			PipeTarget.ToDelegate(NotifyOnStart(out var tcs)),
			PipeTarget.ToStringBuilder(stdout)
		);

		var cmd = new Command(Dummy.Program.FilePath)
			          .WithArguments(["sleep", "00:00:30"])
		          | target;

		// Act
		var task = cmd.ExecuteAsync();
		await tcs.Task;
		task.Kill();

		// Assert
		var exception = Assert.ThrowsAsync<CommandExecutionException>(async () => await task);

		Assert.Multiple(() =>
		{
			Assert.That(ProcessTree.HasExited(task.ProcessId));
			Assert.That(stdout.ToString(), Does.Contain("Sleeping for"));
			Assert.That(stdout.ToString(), Does.Not.Contain("Done."));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_kill_it_immediately()
	{
		// Arrange
		var stdout = new StringBuilder();

		var target = PipeTarget.Merge(
			PipeTarget.ToDelegate(NotifyOnStart(out var tcs)),
			PipeTarget.ToStringBuilder(stdout)
		);

		var cmd = new Command(Dummy.Program.FilePath)
			          .WithArguments(["sleep", "00:00:30"])
		          | target;

		// Act
		var task = cmd.ExecuteBufferedAsync();
		await tcs.Task;
		task.Kill();

		// Assert
		var exception = Assert.ThrowsAsync<CommandExecutionException>(async () => await task);

		Assert.Multiple(() =>
		{
			Assert.That(ProcessTree.HasExited(task.ProcessId));
			Assert.That(stdout.ToString(), Does.Contain("Sleeping for"));
			Assert.That(stdout.ToString(), Does.Not.Contain("Done."));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_and_interrupt_it_immediately()
	{
		// Arrange
		var stdout = new StringBuilder();

		var target = PipeTarget.Merge(
			PipeTarget.ToDelegate(NotifyOnStart(out var tcs)),
			PipeTarget.ToStringBuilder(stdout)
		);

		var cmd = new Command(Dummy.Program.FilePath)
			          .WithArguments(["sleep", "00:00:30"])
		          | target;

		// Act
		var task = cmd.ExecuteAsync();
		await tcs.Task;
		task.Interrupt();

		// Assert
		Assert.DoesNotThrowAsync(async () => await task);

		Assert.Multiple(() =>
		{
			Assert.That(ProcessTree.HasExited(task.ProcessId));
			Assert.That(stdout.ToString(), Does.Contain("Sleeping for"));
			Assert.That(stdout.ToString(), Does.Contain("Done."));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_with_buffering_and_interrupt_it_immediately()
	{
		// Arrange
		var stdout = new StringBuilder();

		var target = PipeTarget.Merge(
			PipeTarget.ToDelegate(NotifyOnStart(out var tcs)),
			PipeTarget.ToStringBuilder(stdout)
		);

		var cmd = new Command(Dummy.Program.FilePath)
			          .WithArguments(["sleep", "00:00:30"])
		          | target;

		// Act
		var task = cmd.ExecuteBufferedAsync();
		await tcs.Task;
		task.Interrupt();

		// Assert
		Assert.DoesNotThrowAsync(async () => await task);

		Assert.Multiple(() =>
		{
			Assert.That(ProcessTree.HasExited(task.ProcessId));
			Assert.That(stdout.ToString(), Does.Contain("Sleeping for"));
			Assert.That(stdout.ToString(), Does.Contain("Done."));
		});
	}
}