// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.ComponentModel;

using Geekeey.Extensions.Process.Buffered;

namespace Geekeey.Extensions.Process.Tests;

[TestFixture]
internal sealed class ExecuteTests
{
	[Test]
	public async Task I_can_execute_a_command_and_get_the_exit_code_and_execution_time()
	{
		// Arrange
		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["echo"]);

		// Act
		var result = await cmd.ExecuteAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.ExitCode, Is.EqualTo(0));
			Assert.That(result.IsSuccess, Is.True);
			Assert.That(result.RunTime, Is.GreaterThan(TimeSpan.Zero));
		});
	}

	[Test]
	public async Task I_can_execute_a_command_and_get_the_associated_process_id()
	{
		// Arrange
		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["echo"]);

		// Act
		var task = cmd.ExecuteAsync();

		// Assert
		Assert.That(task.ProcessId, Is.Not.EqualTo(0));

		await task;
	}

	[Test]
	public async Task I_can_execute_a_command_with_a_configured_awaiter()
	{
		// Arrange
		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["echo"]);

		// Act + Assert
		await cmd.ExecuteAsync().ConfigureAwait(false);
	}

	[Test]
	public Task I_can_try_to_execute_a_command_and_get_an_error_if_the_target_file_does_not_exist()
	{
		// Arrange
		var cmd = new Command("some_exe_with_does_not_exits");

		// Act + Assert
		Assert.ThrowsAsync<Win32Exception>(async () => await cmd.ExecuteAsync());
		return Task.CompletedTask;
	}

	[Test]
	public async Task I_can_execute_a_command_with_a_custom_working_directory()
	{
		// Arrange
		using var dir = TestTempDirectory.Create();

		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments("cwd")
			.WithWorkingDirectory(dir.Path);

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		var lines = result.StandardOutput.Split(Environment.NewLine);
		Assert.That(lines, Is.SupersetOf(new[] { dir.Path }));
	}

	[Test]
	public async Task I_can_execute_a_command_with_additional_environment_variables()
	{
		// Arrange
		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["env", "foo", "bar"])
			.WithEnvironmentVariables(env => env
				.Set("foo", "hello")
				.Set("bar", "world"));

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		var lines = result.StandardOutput.Split(Environment.NewLine);
		Assert.That(lines, Is.SupersetOf(new[] { "hello", "world" }));
	}

	[Test]
	public async Task I_can_execute_a_command_with_some_environment_variables_overwritten()
	{
		// Arrange
		var key = Guid.NewGuid();
		var variableToKeep = $"GKY_TEST_KEEP_{key}";
		var variableToOverwrite = $"GKY_TEST_OVERWRITE_{key}";
		var variableToUnset = $"GKY_TEST_UNSET_{key}";

		using var _a = (TestEnvironment.Create(variableToKeep, "keep"));
		using var _b = (TestEnvironment.Create(variableToOverwrite, "overwrite"));
		using var _c = (TestEnvironment.Create(variableToUnset, "unset"));

		var cmd = new Command(Dummy.Program.FilePath)
			.WithArguments(["env", variableToKeep, variableToOverwrite, variableToUnset])
			.WithEnvironmentVariables(env => env
				.Set(variableToOverwrite, "overwritten")
				.Set(variableToUnset, null));

		// Act
		var result = await cmd.ExecuteBufferedAsync();

		Assume.That(result.ExitCode, Is.EqualTo(0));

		// Assert
		var lines = result.StandardOutput.Split(Environment.NewLine);
		Assert.That(lines, Is.SupersetOf(new[] { "keep", "overwritten" }));
	}
}