// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using Geekeey.Extensions.Process.Buffered;

namespace Geekeey.Extensions.Process.Tests;

[TestFixture]
internal sealed class ValidationTests
{
	private static Command Exit() => new Command(Dummy.Program.FilePath)
		.WithArguments(["exit", "1"]);

	[Test]
	public void I_can_try_to_execute_a_command_and_get_an_error_if_it_returns_a_non_zero_exit_code()
	{
		// Arrange
		var cmd = Exit();

		// Act & Assert
		var exception = Assert.ThrowsAsync<CommandExecutionException>(async () => await cmd.ExecuteAsync());

		Assert.Multiple(() =>
		{
			Assert.That(exception.ExitCode, Is.EqualTo(1));
			Assert.That(exception.Message, Does.Contain("a non-zero exit code (1)"));
		});

		Console.WriteLine(exception.ToString());
	}

	[Test]
	public void I_can_try_to_execute_a_command_with_buffering_and_get_a_detailed_error_if_it_returns_a_non_zero_exit_code()
	{
		// Arrange
		var cmd = Exit();


		// Act & Assert
		var exception = Assert.ThrowsAsync<CommandExecutionException>(async () => await cmd.ExecuteBufferedAsync());

		Assert.Multiple(() =>
		{
			Assert.That(exception.ExitCode, Is.EqualTo(1));
			Assert.That(exception.Message, Does.Contain("Exit code set to 1"));
		});

		Console.WriteLine(exception.ToString());
	}

	[Test]
	public async Task I_can_execute_a_command_without_validating_the_exit_code()
	{
		// Arrange
		var cmd = Exit()
			.WithValidation(CommandExitBehaviour.None);

		// Act
		var result = await cmd.ExecuteAsync();

		// Assert
		Assert.Multiple(() =>
		{
			Assert.That(result.ExitCode, Is.EqualTo(1));
			Assert.That(result.IsSuccess, Is.False);
		});
	}
}