// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.Diagnostics.CodeAnalysis;
using System.Diagnostics.Contracts;

namespace Geekeey.Extensions.Process;

public sealed partial class Command
{
	/// <summary>
	/// Creates a copy of this command, setting the target file path to the specified value.
	/// </summary>
	[Pure]
	public Command WithTargetFile(string targetFilePath)
		=> new(targetFilePath, Arguments, WorkingDirPath, EnvironmentVariables, Validation,
			StandardInputPipe, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the arguments to the specified value.
	/// </summary>
	/// <remarks>
	/// Avoid using this overload, as it requires the arguments to be escaped manually.
	/// Formatting errors may lead to unexpected bugs and security vulnerabilities.
	/// </remarks>
	[Pure]
	public Command WithArguments(string arguments)
		=> new(TargetFilePath, arguments, WorkingDirPath, EnvironmentVariables, Validation,
			StandardInputPipe, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the arguments to the value
	/// obtained by formatting the specified enumeration.
	/// </summary>
	[Pure]
	public Command WithArguments(IEnumerable<string> arguments, bool escape = true)
		=> WithArguments(args => args.Add(arguments, escape));

	/// <summary>
	/// Creates a copy of this command, setting the arguments to the value
	/// configured by the specified delegate.
	/// </summary>
	[Pure]
	public Command WithArguments(Action<ArgumentsBuilder> configure)
	{
		var builder = new ArgumentsBuilder();
		configure(builder);

		return WithArguments(builder.Build());
	}

	/// <summary>
	/// Creates a copy of this command, setting the working directory path to the specified value.
	/// </summary>
	[Pure]
	public Command WithWorkingDirectory(string workingDirPath)
		=> new(TargetFilePath, Arguments, workingDirPath, EnvironmentVariables, Validation,
			StandardInputPipe, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the environment variables to the specified value.
	/// </summary>
	[Pure]
	public Command WithEnvironmentVariables(IReadOnlyDictionary<string, string?> environmentVariables)
		=> new(TargetFilePath, Arguments, WorkingDirPath, environmentVariables, Validation,
			StandardInputPipe, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the environment variables to the value
	/// configured by the specified delegate.
	/// </summary>
	[Pure]
	public Command WithEnvironmentVariables(Action<EnvironmentVariablesBuilder> configure)
	{
		var builder = new EnvironmentVariablesBuilder();
		configure(builder);

		return WithEnvironmentVariables(builder.Build());
	}

	/// <summary>
	/// Creates a copy of this command, setting the validation options to the specified value.
	/// </summary>
	[Pure]
	public Command WithValidation(CommandExitBehaviour exitBehaviour)
		=> new(TargetFilePath, Arguments, WorkingDirPath, EnvironmentVariables, exitBehaviour,
			StandardInputPipe, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the standard input pipe to the specified source.
	/// </summary>
	[Pure]
	public Command WithStandardInputPipe(PipeSource source)
		=> new(TargetFilePath, Arguments, WorkingDirPath, EnvironmentVariables, Validation,
			source, StandardOutputPipe, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the standard output pipe to the specified target.
	/// </summary>
	[Pure]
	public Command WithStandardOutputPipe(PipeTarget target)
		=> new(TargetFilePath, Arguments, WorkingDirPath, EnvironmentVariables, Validation,
			StandardInputPipe, target, StandardErrorPipe);

	/// <summary>
	/// Creates a copy of this command, setting the standard error pipe to the specified target.
	/// </summary>
	[Pure]
	public Command WithStandardErrorPipe(PipeTarget target)
		=> new(TargetFilePath, Arguments, WorkingDirPath, EnvironmentVariables, Validation,
			StandardInputPipe, StandardOutputPipe, target);

	/// <inheritdoc />
	[ExcludeFromCodeCoverage]
	public override string ToString() => $"{TargetFilePath} {Arguments}";
}