// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.Diagnostics;

namespace Geekeey.Extensions.Process;

public partial class Command
{
	private static readonly Lazy<string?> ProcessPathLazy = new(() =>
	{
		using var process = System.Diagnostics.Process.GetCurrentProcess();
		return process.MainModule?.FileName;
	});

	private static readonly string[] WindowsExecutableExtensions = ["exe", "cmd", "bat"];
	private static readonly TimeSpan CancelWaitTimeout = TimeSpan.FromSeconds(5);

	private static string? ProcessPath => ProcessPathLazy.Value;

	private ProcessStartInfo CreateStartInfo()
	{
		var info = new ProcessStartInfo
		{
			FileName = GetOptimallyQualifiedTargetFilePath(),
			Arguments = Arguments,
			WorkingDirectory = WorkingDirPath,
			RedirectStandardInput = true,
			RedirectStandardOutput = true,
			RedirectStandardError = true,
			UseShellExecute = false,
			// This option only works on Windows and is required there to prevent the
			// child processes from attaching to the parent console window, if one exists.
			// We need this in order to be able to send signals to one specific child process,
			// without affecting any others that may also be running in parallel.
			CreateNoWindow = true
		};

		// Set environment variables
		foreach (var (key, value) in EnvironmentVariables)
		{
			if (value is not null)
			{
				info.Environment[key] = value;
			}
			else
			{
				// Null value means we should remove the variable
				info.Environment.Remove(key);
			}
		}

		return info;

		string GetOptimallyQualifiedTargetFilePath()
		{
			// Currently, we only need this workaround for script files on Windows, so short-circuit
			// if we are on a different platform.
			if (!OperatingSystem.IsWindows())
				return TargetFilePath;

			// Don't do anything for fully qualified paths or paths that already have an extension specified.
			// System.Diagnostics.Process knows how to handle those without our help.
			if (Path.IsPathFullyQualified(TargetFilePath) ||
			    !string.IsNullOrWhiteSpace(Path.GetExtension(TargetFilePath)))
				return TargetFilePath;

			return (
				from probeDirPath in GetProbeDirectoryPaths()
				where Directory.Exists(probeDirPath)
				select Path.Combine(probeDirPath, TargetFilePath)
				into baseFilePath
				from extension in WindowsExecutableExtensions
				select Path.ChangeExtension(baseFilePath, extension)
			).FirstOrDefault(File.Exists) ?? TargetFilePath;

			static IEnumerable<string> GetProbeDirectoryPaths()
			{
				// Executable directory
				if (!string.IsNullOrWhiteSpace(ProcessPath))
				{
					var processDirPath = Path.GetDirectoryName(ProcessPath);

					if (!string.IsNullOrWhiteSpace(processDirPath))
						yield return processDirPath;
				}

				// Working directory
				yield return Directory.GetCurrentDirectory();

				// Directories on the PATH
				if (Environment.GetEnvironmentVariable("PATH")?.Split(Path.PathSeparator) is { } paths)
				{
					foreach (var path in paths)
					{
						yield return path;
					}
				}
			}
		}
	}

	/// <summary>
	/// Executes the command asynchronously.
	/// </summary>
	/// <remarks>
	/// This method can be awaited.
	/// </remarks>
	public CommandTask<CommandResult> ExecuteAsync(CancellationToken cancellationToken = default)
	{
		var process = new ProcessHandle(CreateStartInfo());
		process.Start();

		// Extract the process ID before calling ExecuteAsync(), because the process may already be disposed by then.
		var processId = process.Id;

		var task = ExecuteAsync(process, cancellationToken);
		return new CommandTask<CommandResult>(task, process, processId);
	}

	private async Task<CommandResult> ExecuteAsync(ProcessHandle process, CancellationToken cancellationToken = default)
	{
		using var _ = process;

		// Additional cancellation for the stdin pipe in case the process exits without fully exhausting it
		using var stdin = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
		using var complete = new CancellationTokenSource();

		// ReSharper disable once AccessToDisposedClosure
		await using (cancellationToken.Register(() => complete.CancelAfter(CancelWaitTimeout)))
		await using (cancellationToken.Register(process.Kill))
		{
			var writing = PipeStdInAsync(process, stdin.Token);

			var reading = Task.WhenAll(
				PipeStdOutAsync(process, cancellationToken),
				PipeStdErrAsync(process, cancellationToken));

			// Wait until the process exits normally or gets killed. The timeout is started after the execution is
			// forcefully canceled and ensures that we don't wait forever in case the attempt to kill the process failed.
			await process.WaitUntilExitAsync(complete.Token);

			// Send the cancellation signal to the stdin pipe since the process has exited and won't need it anymore.
			// If the pipe has already been exhausted (most likely), this won't do anything. If the pipe is still trying
			// to transfer data, this will cause it to abort.
			await stdin.CancelAsync();

			// Wait until piping is done and propagate exceptions.
			await Task.WhenAll(writing, reading);
		}

		if (process.ExitCode is 0 || !Validation.HasFlag(CommandExitBehaviour.ZeroExitCode))
		{
			return new CommandResult(process.ExitCode, process.StartTime, process.ExitTime);
		}

		var message = $"Command execution failed because the underlying process ({process.Name}#{process.Id}) " +
		              $"returned a non-zero exit code ({process.ExitCode}).";
		throw new CommandExecutionException(this, process.ExitCode, message);
	}

	private async Task PipeStdOutAsync(ProcessHandle process, CancellationToken cancellationToken = default)
	{
		await using (process.StandardOutput)
		{
			await StandardOutputPipe.CopyFromAsync(process.StandardOutput, cancellationToken);
		}
	}

	private async Task PipeStdErrAsync(ProcessHandle process, CancellationToken cancellationToken = default)
	{
		await using (process.StandardError)
		{
			await StandardErrorPipe.CopyFromAsync(process.StandardError, cancellationToken);
		}
	}

	private async Task PipeStdInAsync(ProcessHandle process, CancellationToken cancellationToken = default)
	{
		await using (process.StandardInput)
		{
			try
			{
				// Some streams do not support cancellation, so we add a fallback that drops the task and returns early.
				// This is important with stdin because the process might finish before the pipe has been fully
				// exhausted, and we don't want to wait for it.
				await StandardInputPipe.CopyToAsync(process.StandardInput, cancellationToken)
					.WaitAsync(cancellationToken);
			}
			// Expect IOException: "The pipe has been ended" (Windows) or "Broken pipe" (Unix). This may happen if the
			// process terminated before the pipe has been exhausted. It's not an exceptional situation because the
			// process may not need the entire stdin to complete successfully. We also can't rely on process.HasExited
			// here because of potential race conditions.
			catch (IOException ex) when (ex.GetType() == typeof(IOException))
			{
				// Don't catch derived exceptions, such as FileNotFoundException, to avoid false positives.
			}
		}
	}
}