﻿// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

namespace Geekeey.Extensions.Process;

/// <summary>
/// Instructions for running a process.
/// </summary>
public sealed partial class Command(
	string targetFilePath,
	string arguments,
	string workingDirPath,
	IReadOnlyDictionary<string, string?> environmentVariables,
	CommandExitBehaviour validation,
	PipeSource standardInputPipe,
	PipeTarget standardOutputPipe,
	PipeTarget standardErrorPipe
)
{
	/// <summary>
	/// Initializes an instance of <see cref="Command" />.
	/// </summary>
	public Command(string targetFilePath) : this(targetFilePath, string.Empty, Directory.GetCurrentDirectory(),
		new Dictionary<string, string?>(),
		CommandExitBehaviour.ZeroExitCode, PipeSource.Null, PipeTarget.Null, PipeTarget.Null)
	{
	}

	/// <summary>
	/// File path of the executable, batch file, or script, that this command runs.
	/// </summary>
	public string TargetFilePath { get; } = targetFilePath;

	/// <summary>
	/// File path of the executable, batch file, or script, that this command runs.
	/// </summary>
	public string Arguments { get; } = arguments;

	/// <summary>
	/// File path of the executable, batch file, or script, that this command runs.
	/// </summary>
	public string WorkingDirPath { get; } = workingDirPath;

	/// <summary>
	/// Environment variables set for the underlying process.
	/// </summary>
	public IReadOnlyDictionary<string, string?> EnvironmentVariables { get; } = environmentVariables;

	/// <summary>
	/// Strategy for validating the result of the execution.
	/// </summary>
	public CommandExitBehaviour Validation { get; } = validation;

	/// <summary>
	/// Pipe source for the standard input stream of the underlying process.
	/// </summary>
	public PipeSource StandardInputPipe { get; } = standardInputPipe;

	/// <summary>
	/// Pipe target for the standard output stream of the underlying process.
	/// </summary>
	public PipeTarget StandardOutputPipe { get; } = standardOutputPipe;

	/// <summary>
	/// Pipe target for the standard error stream of the underlying process.
	/// </summary>
	public PipeTarget StandardErrorPipe { get; } = standardErrorPipe;
}