// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.ComponentModel;
using System.Diagnostics;

namespace Geekeey.Extensions.Process;

internal sealed partial class ProcessHandle(ProcessStartInfo startInfo) : IDisposable
{
	private readonly System.Diagnostics.Process _process = new() { StartInfo = startInfo };
	private readonly TaskCompletionSource<object?> _exitTcs = new(TaskCreationOptions.RunContinuationsAsynchronously);

	public int Id => _process.Id;

	// we have to keep track of ProcessName ourselves because it becomes inaccessible after the process exits
	public string Name => Path.GetFileName(_process.StartInfo.FileName);

	// we are purposely using Stream instead of StreamWriter/StreamReader to push the concerns of
	// writing and reading to PipeSource/PipeTarget at the higher level.
	public Stream StandardInput => _process.StandardInput.BaseStream;

	public Stream StandardOutput => _process.StandardOutput.BaseStream;

	public Stream StandardError => _process.StandardError.BaseStream;

	// we have to keep track of StartTime ourselves because it becomes inaccessible after the process exits
	public DateTimeOffset StartTime { get; private set; }

	// we have to keep track of ExitTime ourselves because it becomes inaccessible after the process exits
	public DateTimeOffset ExitTime { get; private set; }

	public int ExitCode => _process.ExitCode;

	public void Start()
	{
		_process.EnableRaisingEvents = true;
		_process.Exited += (_, _) =>
		{
			ExitTime = DateTimeOffset.Now;
			_exitTcs.TrySetResult(null);
		};

		try
		{
			if (!_process.Start())
			{
				throw new InvalidOperationException(
					$"Failed to start a process with file path '{_process.StartInfo.FileName}'. " +
					$"Target file is not an executable or lacks execute permissions.");
			}

			StartTime = DateTimeOffset.Now;
		}
		catch (Win32Exception exception)
		{
			throw new Win32Exception(
				$"Failed to start a process with file path '{_process.StartInfo.FileName}'. " +
				$"Target file or working directory doesn't exist, or the provided credentials are invalid.", exception);
		}
	}

	public void Interrupt()
	{
		if (TryInterrupt()) return;

		// In case of failure, revert to the default behavior of killing the process.
		// Ideally, we should throw an exception here, but this method is called from
		// a cancellation callback, which would prevent other callbacks from being called.
		Kill();
		Debug.Fail("Failed to send an interrupt signal.");

		return;

		bool TryInterrupt()
		{
			try
			{
				if (OperatingSystem.IsWindows())
				{
					return SendCtrlSignal(_process.Id, ConsoleCtrlEvent.CTRL_C_EVENT) ||
					       SendCtrlSignal(_process.Id, ConsoleCtrlEvent.CTRL_BREAK_EVENT);
				}

				if (OperatingSystem.IsLinux() || OperatingSystem.IsMacOS() || OperatingSystem.IsFreeBSD())
				{
					return SendPosixSignal(_process.Id, PosixSignals.SIGINT) ||
					       SendPosixSignal(_process.Id, PosixSignals.SIGTERM);
				}

				// Unsupported platform
				return false;
			}
			catch
			{
				return false;
			}
		}
	}

	public void Kill()
	{
		try
		{
			_process.Kill(true);
		}
		catch when (_process.HasExited)
		{
			// The process has exited before we could kill it. This is fine.
		}
		catch
		{
			// The process either failed to exit or is in the process of exiting.
			// We can't really do anything about it, so just ignore the exception.
			Debug.Fail("Failed to kill the process.");
		}
	}

	public async Task WaitUntilExitAsync(CancellationToken cancellationToken = default)
	{
		await _exitTcs.Task.WaitAsync(cancellationToken);
	}

	public void Dispose() => _process.Dispose();
}