// Copyright (c) The Geekeey Authors
// SPDX-License-Identifier: EUPL-1.2

using System.Buffers;
using System.Diagnostics.CodeAnalysis;

namespace Geekeey.Extensions.Process;

internal sealed class MemoryBufferStream : Stream
{
	private readonly SemaphoreSlim _writeLock = new(1, 1);
	private readonly SemaphoreSlim _readLock = new(0, 1);

	private IMemoryOwner<byte> _sharedBuffer = MemoryPool<byte>.Shared.Rent(BufferSizes.Stream);
	private int _sharedBufferBytes;
	private int _sharedBufferBytesRead;

	[ExcludeFromCodeCoverage] public override bool CanRead => true;

	[ExcludeFromCodeCoverage] public override bool CanSeek => false;

	[ExcludeFromCodeCoverage] public override bool CanWrite => true;

	[ExcludeFromCodeCoverage] public override long Position { get; set; }

	[ExcludeFromCodeCoverage] public override long Length => throw new NotSupportedException();

	[ExcludeFromCodeCoverage]
	public override void Write(byte[] buffer, int offset, int count)
		=> WriteAsync(buffer, offset, count).GetAwaiter().GetResult();

	[ExcludeFromCodeCoverage]
	public override async Task WriteAsync(byte[] buffer, int offset, int count, CancellationToken cancellationToken)
		=> await WriteAsync(buffer.AsMemory(offset, count), cancellationToken);

	public override async ValueTask WriteAsync(ReadOnlyMemory<byte> buffer,
		CancellationToken cancellationToken = default)
	{
		await _writeLock.WaitAsync(cancellationToken);

		// Reset the buffer if the current one is too small for the incoming data
		if (_sharedBuffer.Memory.Length < buffer.Length)
		{
			_sharedBuffer.Dispose();
			_sharedBuffer = MemoryPool<byte>.Shared.Rent(buffer.Length);
		}

		buffer.CopyTo(_sharedBuffer.Memory);

		_sharedBufferBytes = buffer.Length;
		_sharedBufferBytesRead = 0;

		_readLock.Release();
	}

	[ExcludeFromCodeCoverage]
	public override int Read(byte[] buffer, int offset, int count)
		=> ReadAsync(buffer, offset, count).GetAwaiter().GetResult();

	[ExcludeFromCodeCoverage]
	public override async Task<int>
		ReadAsync(byte[] buffer, int offset, int count, CancellationToken cancellationToken) =>
		await ReadAsync(buffer.AsMemory(offset, count), cancellationToken);

	public override async ValueTask<int> ReadAsync(Memory<byte> buffer, CancellationToken cancellationToken = default)
	{
		await _readLock.WaitAsync(cancellationToken);

		var length = Math.Min(buffer.Length, _sharedBufferBytes - _sharedBufferBytesRead);

		_sharedBuffer.Memory.Slice(_sharedBufferBytesRead, length).CopyTo(buffer);

		_sharedBufferBytesRead += length;

		// release the write lock if the consumer has finished reading all
		// the previously written data.
		if (_sharedBufferBytesRead >= _sharedBufferBytes)
		{
			_writeLock.Release();
		}
		// otherwise, release the read lock again so that the consumer can finish
		// reading the data.
		else
		{
			_readLock.Release();
		}

		return length;
	}

	public async Task ReportCompletionAsync(CancellationToken cancellationToken = default) =>
		// write an empty buffer that will make ReadAsync(...) return 0, which signals the end of stream
		await WriteAsync(Memory<byte>.Empty, cancellationToken);

	protected override void Dispose(bool disposing)
	{
		if (disposing)
		{
			_readLock.Dispose();
			_writeLock.Dispose();
			_sharedBuffer.Dispose();
		}

		base.Dispose(disposing);
	}

	[ExcludeFromCodeCoverage]
	public override void Flush() => throw new NotSupportedException();

	[ExcludeFromCodeCoverage]
	public override long Seek(long offset, SeekOrigin origin) => throw new NotSupportedException();

	[ExcludeFromCodeCoverage]
	public override void SetLength(long value) => throw new NotSupportedException();
}